/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
*
* @FILE     li_async.c
*
* @ABSTRACT
* This file contains:
*
* @INCLUDE
*/


#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <time.h>
#include <sys/time.h>
#include <unistd.h>

#include <cmqc.h>

#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "gcn_upext.h"
#include "gcn_upmac.h"
#include "gcn_upmes.h"

#include "osx_upcon.h"
#include "osx_upext.h"

#include <cdx_upcon.h>
#include "cdx_uptyp.h"
#include "cdx_upext.h"
#include "cdx_upmac.h"

#include "prs_upcon.h"
#include "prs_upext.h"

#include "csi_upcon.h"
#include "csi_uptyp.h"
#include "csi_usext.h"

#include "rou_upcon.h"
#include "rou_upext.h"

#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "evr_usmes.h"
#include "li_prmes.h"

#include "li_prcon.h"
#include "li_prtyp.h"
#include "li_prext.h"
#include "li_upext.h"

/**************************************************************/
/*** THE LOCAL INTELLIGENCE ASYNCHRONOUS PROCESSING ROUTINE ***/
/**************************************************************/
/*
* @FUNCTION li_AsyncProc
*
* @DESCRIPTION Processes the asynchronous messages in the LI queue.
*
* @ABSTRACT
* The Local Intelligence Asynchronous Processing routine is called by
* the <@FUNCTION li_service> upon reception of an asynchronous message.
*
* Actually, the LI is called by the Trigger Monitor; the name of the queue
* is transmitted with that call (which we call a Trigger Message). The Trigger
* Message has already been processed, and we now enter a "retrieving loop".
* 
* This function:
* - opens the queue with <@FUNCTION CSI_mq_open>
* - gets a message in it with <@FUNCTION CSI_mq_get> and processes it until
*   the queue is empty (which is indicated by the reason code 
*   MQ_ERR_NO_MSG_AVAILABLE of the 'get').
*
* The processing of an asynchronous message:
* An asynchronous message may be:
* - a quite normal asynchronous message, namely either a datagram, a request
*   or a reply (for CCN/CSI, those three types are equivalent).
* - a degraded mode message.
* - a report.
*
* Each type of asynchronous message supposes a different treatment; among the
* reports, there are even different kinds (user-, MQS- or CCN-generated).
*
* For a "normal" asynchronous message, the processing is:
* - extracting trace info with <@FUNCTION gcn_GetTraceInfo>.
* - calling the operation dispatcher which performs the different tasks, 
*   i.e. routing, access control, presentation, VAS with
*   <@LOCALFUN li_OperationDispatcher>.
* - rebuilding trace info with <@FUNCTION gcn_BuildTraceInfo>.
* - putting the message in the "next hop" queue with <@FUNCTION CSI_mq_put1>.
*
* @RETURNVALUE CSILONG
* Possible returnvalues: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER 
* IN     LI_PROCCONTEXT *li_ProcContext_P
*        A pointer to the LI process context.
*
* INOUT  LI_MSGCONTEXT *li_MsgContext_P
*        A pointer to the LI current message context.
*
* OUT    CSILONG *returnCode_P
*        Possible values CSISUCCESS or CSIFAILED.
*
* OUT    CSILONG *reasonCode_P
*        Reason Code.
*
* @SEEALSO
* 
*/
CSILONG   li_AsyncProc(
IN        LI_PROCCONTEXT  *li_ProcContext_P,
INOUT     LI_MSGCONTEXT   *li_MsgContext_P,
OUT       CSILONG         *returnCode_P,
OUT       CSILONG         *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   /** Local variables associated with MQSeries **/

   CSIMQOD     objDesc_S       = {CSIMQOD_DEFAULT};

   MQLONG       attrSelectorSet[ALI_MQSET_REQUESTED_ATTR_MXNUM]; /* MQSET */
   MQLONG       attrValuesSet[ALI_MQSET_REQUESTED_ATTR_MXNUM];   /* MQSET */

   CSILONG      mqResultCode;
   CSILONG      mqReasonCode;

   CSILONG      ccnVersion;
   CSICHAR48    tt2;
   CSICHAR48    tt3;
   CSICHAR256 szServiceDN;

   /** Reminder: no need to declare a CSIMQMD: there is already one in **/
   /**           the LI message context.                               **/

   CSIMQHCONN  objHandle       = 0L;   /* Object (queue) handle */

   CSILONG     o_Options       = 0L;   /* Open  options */
   CSILONG     c_Options       = 0L;   /* Close options */

   CSIPDU     *msgDataDesc_P   = NULL;

   CSILONG     msgLen          = 0L;

   CSILONG     ret             = CSISUCCESS;       /* All purpose ret var */
   CSILONG     status          = LI_ERR_NONE;      /* Error status        */

   CSIBOOLEAN  open_ok         = CSIFALSE;         /* Is the Q open ?     */

   TRACE_FUNCTION_ENTRY("li_AsyncProc");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ((NULL==returnCode_P)||(NULL==reasonCode_P))
   {
      TRACE_FUNCTION_EXIT("li_AsyncProc",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /*** OR4237: CONNECTION TO THE LOCAL QUEUE MANAGER ***/
   /*****************************************************/
   /** As no queue manager name is mentioned, the default local queue **/
   /** manager is selected.                                           **/

   if (LI_ERR_NONE == status)
   {
      ret = CSI_mq_conn("",&(li_ProcContext_P->QMgrHandle),returnCode_P,reasonCode_P);
      if (CSIFAILED == ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
         "CSI_mq_conn",(*reasonCode_P)));
         status = *reasonCode_P;

        li_ProcContext_P->QMgrHandle = CSIMQHC_UNUSABLE_HCONN;
      }
   }


   /*** OPENING THE TRIGGERED QUEUE ***/
   /***********************************/

   if (LI_ERR_NONE == status)
   {
      /* o_Options = CSIMQOO_INPUT_AS_Q_DEF + CSIMQOO_FAIL_IF_QUIESCING; */
      o_Options = CSIMQOO_INPUT_AS_Q_DEF + CSIMQOO_FAIL_IF_QUIESCING + CSIMQOO_SET ;

      strncpy(objDesc_S.ObjectName,li_MsgContext_P->TriggerQName,
              sizeof(CSICHAR48));

      ret = CSI_mq_open(li_ProcContext_P->QMgrHandle,&objDesc_S, o_Options,
                        &objHandle,returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         open_ok = CSIFALSE;
         EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_TRGQ_OPEN_ERROR,
         li_MsgContext_P->TriggerQName,(*reasonCode_P)));
         status = *reasonCode_P;
      }
      else
      {
         open_ok = CSITRUE;
      }
   }

   /*** RETRIEVING LOOP IN THE TRIGGERING QUEUE ***/
   /***********************************************/
   /********************************************************************/
   /* Note: the loop goes on as long as the CSI_mq_get is successfully */
   /* ----- performed, meaning that if it is not performed, or it is   */
   /*       performed but fails, the loop stops.                       */
   /*       The booleans get_done and get_ok trace this behaviour.     */
   /********************************************************************/
   
   while (LI_ERR_NONE == status)
   { /** BEGIN: RETRIEVING LOOP **/

      /** LOOP VARIABLES DECLARATION AND INITIALISATION **/
      /***************************************************/

      CSIMQGMO    getMsgOpt_S     = {CSIMQGMO_DEFAULT}; /* 'get' options struct */
      CSIMQPMO    putMsgOpt_S     = {CSIMQPMO_DEFAULT}; /* 'put' options struct */

      CSILONG     toLocal         = CSITRUE;            /* To local gateway ?   */
      CSILONG     opFlags         = LI_DEFAULT_FLAGS;   /* Op.Dispatcher flags  */

      CSILONG     dead_queue      = CSIFALSE;           /* msg put in DLQ ?     */

      CSILONG     get_done        = CSIFALSE;           /* 'get' performed ?    */
      CSILONG     get_ok          = CSIFALSE;           /* 'get' successful ?   */

      CSILONG     interface       = ROU_LI_TO_EXT;      /* For report creation  */

      CSICHAR48   deadLetterQName = {""}; 

      CSICHAR256  szReportFileName = {""};              /* for a symbolic link  */

      /* Audit purposes */

      CSILONG     tmpRet          = 0L;   
      CSILONG     tmpReas         = 0L;
      CSIBOOLEAN  isOrigLI        = CSITRUE;
      CSILONG     lMsgLengthBeforeConversion = -1L;

      /* Reports with data */
     CSIBOOLEAN  cancelPutReport = CSIFALSE;

      /* Must a rollback be taken ? */
      CSIBOOLEAN  rollbackTransaction = CSIFALSE;

      /* fields used to store the 'I' audit values */
      CSILONG lSubmissionTime_I = 0;
      CSILONG lSubmissionTimeU_I = 0;
      CSICHAR8 szPeerGateway_I;
      CSICHAR48 szUniqueId_I;
      CSICHAR32 szMsgTypeId_I;
      CSILONG lMsgLength_I = 0;
      CSILONG lReasonCode_I = LI_ERR_NONE;
      CSICHAR48 szUserName_I;
      CSICHAR48 szAppName_I;
      CSICHAR48 szQueueName_I;
      CSILONG lMsgType_I = 0;
      CSIBYTE24 szMsgId_I;
      CSIBYTE24 szCorrelId_I;
      CSILONG lFeedback_I = 0;
      CSICHAR48 szReplyToQueue_I;
      CSILONG lReportRequest_I = CSIMQRO_NONE;     

      li_MsgContext_P->MsgTypeId[0]=0;

      /*** ALLOCATION OF THE msgDataDesc CSIDD, PREPARATIONS ***/
      /*********************************************************/
      /** Warning: this CSIPDU is allocated especially for the GET, verify **/
      /** if the underlying layers are able to reallocate it if needed.   **/
    
      /** Releasing any argument buffer, so that no conflict will appear **/
      /** with those of any previously retrieved asynchronous message.   **/
    
      /** Releasing here the argument buffer containing the UserProfiles **/
    
      if (LI_ERR_NONE == status)
      {
         if (NULL != li_MsgContext_P->UserProfileList_P)
         {
            ret = OS_mem_free((CSIVOID **)&(li_MsgContext_P->UserProfileList_P),
                  returnCode_P, reasonCode_P);
            if (CSIFAILED==ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
               "OS_mem_free",(*reasonCode_P)));
               status = LI_ERR_FREE;
            }
         }
      }


      /** Releasing here the argument buffer containing the Prev.Gateways **/

      if (LI_ERR_NONE == status)
      {
         if (NULL != li_MsgContext_P->PrevGatewayList_P)
         {
            ret = OS_mem_free((CSIVOID **)&(li_MsgContext_P->PrevGatewayList_P),
                  returnCode_P, reasonCode_P);
            if (CSIFAILED==ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
               "OS_mem_free",(*reasonCode_P)));
               status = LI_ERR_FREE;
            }
         }
      }


      /** Releasing the server host format which can be different for
          each asynchronous message and Allocating a new data descriptor **/

      if (LI_ERR_NONE == status)
      {
         li_MsgContext_P->ServerHostFormat[0] = '\0';
         ret = gcn_AllocDataDesc(0L, &msgDataDesc_P, returnCode_P, reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_AllocDataDesc (msgDataDesc_P)",(*reasonCode_P)));
            status        = LI_ERR_ALLOC;
         }
      }

      /*** RETRIEVAL OF THE MESSAGE IN THE QUEUE ***/
      /*********************************************/

      if (LI_ERR_NONE == status)
      {
         getMsgOpt_S.Options = CSIMQGMO_WAIT + 
                               CSIMQGMO_FAIL_IF_QUIESCING + 
                               CSIMQGMO_SYNCPOINT;

         getMsgOpt_S.WaitInterval = CSI_MQGET_TIMEOUT;

         /* Reinitialization of MsgId and CorrelId, to get ANY message */

         memset( li_MsgContext_P->MsgDesc_P->MsgId   , 0, sizeof(CSIBYTE24) );
         memset( li_MsgContext_P->MsgDesc_P->CorrelId, 0, sizeof(CSIBYTE24) );

         ret = CSI_mq_get(li_ProcContext_P->QMgrHandle, objHandle,
                          li_MsgContext_P->MsgDesc_P,&getMsgOpt_S, 
                          (CSIDD *)msgDataDesc_P,
                          &msgLen, li_MsgContext_P->QoS_P, returnCode_P,
                          reasonCode_P);

         /* The 'GET' has been performed, so get_done is updated */

         get_done = CSITRUE;

         if (CSIFAILED==ret)
         {
            if (*reasonCode_P != MQ_ERR_NO_MSG_AVAILABLE)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
                            "CSI_mq_get",(*reasonCode_P)));
            }
            else
            {
               EVR_TRACELOG(EVR_DEBUG,(LI_TRC_CSI_NO_MSG));
            }
            status = *reasonCode_P;
         }
         else
         {
            /* Moreover the 'GET' is OK, so get_ok is updated */
            get_ok = CSITRUE;

            /* retreive the runmode from the CSIDD for AUDIT purpose*/
            switch (msgDataDesc_P->Flags & CSI_DD_FLAGS_MODE_SET)
            {
               case CSI_DD_FLAGS_PRODUCTION_MSG:
                  strcpy(li_MsgContext_P->RunMode,CDX_C_PRODUCTION_STR);
                  break;

               case CSI_DD_FLAGS_BACKUP_MSG:
                  strcpy(li_MsgContext_P->RunMode,CDX_C_BACKUP_STR);
                  break;

               case CSI_DD_FLAGS_TEST_MSG:
                  strcpy(li_MsgContext_P->RunMode,CDX_C_TEST_STR);
                  break;

               case CSI_DD_FLAGS_MAINTENANCE_MSG:
                  strcpy(li_MsgContext_P->RunMode,CDX_C_MAINTENANCE_STR);
                  break;
            }
         }
      }

      /*** EXTRACTION OF THE TRACE INFORMATION OF THE MESSAGE !!! ***/
      /**************************************************************/

      if (LI_ERR_NONE == status)
      {
         ret = gcn_GetTraceInfo(GCN_GET,msgDataDesc_P,
               &(li_MsgContext_P->EventType),
               li_MsgContext_P->TraceInfo_P,NULL,
               &(li_MsgContext_P->UserProfileList_P),
               &(li_MsgContext_P->PrevGatewayList_P),
               returnCode_P, reasonCode_P );
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_GetTraceInfo",(*reasonCode_P)));
            status              = LI_ERR_GCN_TRACE_INFO;
         }

         /* Trace info removed from message */
         li_MsgContext_P->isTrcInfoInMsg = CSIFALSE;
      }

      /* For audit, later */

      isOrigLI = ((li_MsgContext_P->EventType == CSI_LOCAL_EVENT) ? CSITRUE :
                                                                    CSIFALSE);

      /*** IS IT A REPORT MESSAGE ? ***/
      /********************************/

      if (LI_ERR_NONE == status)
      { /** BEGIN: IF OK RM 1 **/

      if (CSIMQMT_REPORT == li_MsgContext_P->MsgDesc_P->MsgType)
      {
         if (CSI_USER_GENERATED == li_MsgContext_P->TraceInfo_P->GenerationMethod)
         {
            if (CSIMQRO_NONE==li_MsgContext_P->QoS_P->ReportRequest)
            {
               /** It is an external user-generated report   **/
               /** It will be considered as a normal message **/
               /** No particular treatment, thus.            **/
            } 
            else
            {
               /** It is an automatically generated report **/
               /** The trace info must be corrected.       **/

               li_MsgContext_P->TraceInfo_P->GenerationMethod =
                  CSI_SYSTEM_GENERATED;
               
               if (0!=li_MsgContext_P->TraceInfo_P->NbOfPrevGateways)
               {
                  li_MsgContext_P->TraceInfo_P->NbOfPrevGateways = 0;
                  ret = OS_mem_free(
                        (CSIVOID **)&(li_MsgContext_P->PrevGatewayList_P),
                        returnCode_P,reasonCode_P);
                  if (CSIFAILED==ret)
                  {
                     EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
                     "OS_mem_free",(*reasonCode_P)));
                     status = LI_ERR_FREE;
                  }
               }

               strncpy(li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName,
                       li_MsgContext_P->QoS_P->ReplyToQ,
                       sizeof(CSICHAR48) );

               memcpy(li_MsgContext_P->QoS_P->CorrelId,
                      li_MsgContext_P->MsgDesc_P->CorrelId,
                      sizeof(CSIBYTE24) );

               li_MsgContext_P->EventType = CSI_LOCAL_EVENT;

               /* Data Truncation (no more data in a system report) */

               /* OR 4790 : Remove data if EXCEPTION  report without data */
               /*                          EXPIRATION                     */
               /* Note that these reports are mutually exclusive          */
               if (
                      (li_MsgContext_P->MsgDesc_P->Feedback == CSIMQFB_EXPIRATION &&
                       (0L == (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DATA_MASK & \
                          (CSIMQRO_EXPIRATION_WITH_FULL_DATA))))
                   ||
                      (li_MsgContext_P->MsgDesc_P->Feedback != CSIMQFB_COD &&
                       li_MsgContext_P->MsgDesc_P->Feedback != CSIMQFB_COA &&
                       li_MsgContext_P->MsgDesc_P->Feedback != CSIMQFB_EXPIRATION &&
                       (0L == (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DATA_MASK & \
                          (CSIMQRO_EXCEPTION_WITH_FULL_DATA))))
                  )
               {
                /* Delete the data if the received report is an
                    expiration or exception report with data. */
                 gcn_DeleteFile(msgDataDesc_P, &tmpRet, &tmpReas);
                 memset(msgDataDesc_P->FileName, '\0', sizeof(msgDataDesc_P->FileName));
                 msgDataDesc_P->DataLength = 0L;
                 CSI_DD_FLAGS_SET_STORAGE(msgDataDesc_P,
                                          CSI_DD_FLAGS_DATA_BY_MEMORY);
               }
               /* end OR 4790 */

               if (((li_MsgContext_P->MsgDesc_P->Feedback == CSIMQFB_COA) ||
                    (li_MsgContext_P->MsgDesc_P->Feedback == CSIMQFB_COD)) &&
                    (li_MsgContext_P->QoS_P->AppliedQoS & CSI_QOS_DATA_SAVED)) /* OR4280 */
               { /* OR 4160 Data truncation for COA or COD report */
                  msgDataDesc_P->DataLength = 0L;
                  memset(msgDataDesc_P->FileName, '\0', sizeof(msgDataDesc_P->FileName));
                  CSI_DD_FLAGS_SET_STORAGE(msgDataDesc_P, 
                                           CSI_DD_FLAGS_DATA_BY_MEMORY);
               }

               if ((li_MsgContext_P->MsgDesc_P->Feedback == CSIMQFB_COA) &&
                   (0L == (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DATA_MASK & \
                      (CSIMQRO_COA_WITH_FULL_DATA))))
               {
                  msgDataDesc_P->DataLength = 0L;
                   memset(msgDataDesc_P->FileName, '\0', sizeof(msgDataDesc_P->FileName));
                  CSI_DD_FLAGS_SET_STORAGE(msgDataDesc_P, 
                                           CSI_DD_FLAGS_DATA_BY_MEMORY);
               }

               if ((li_MsgContext_P->MsgDesc_P->Feedback == CSIMQFB_COD) &&
                   (0L == (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DATA_MASK & \
                      (CSIMQRO_COD_WITH_FULL_DATA))))
               {
                  msgDataDesc_P->DataLength = 0L;
                  memset(msgDataDesc_P->FileName, '\0', sizeof(msgDataDesc_P->FileName));
                  CSI_DD_FLAGS_SET_STORAGE(msgDataDesc_P, 
                                           CSI_DD_FLAGS_DATA_BY_MEMORY);
               }

               /** No ACM, nor VAS, nor Data Conversion for a report **/
               /** generated by the CCN/CSI system (auto or not).    **/

               /* OR 4148 : checking ACM for report now */
               /* OR 4458 : data conversion for report now */
               opFlags = /* LI_NO_ACM + */ LI_NO_VAS /* + LI_NO_MSGCONV */;
            }
         }
         else
         { /* It is a CCN system report (sent by a technical server) */
   
            /* OR 4148 : checking ACM for report now */ 
            /* OR 4458 : data conversion for report now */
            opFlags = /* LI_NO_ACM  + */ LI_NO_VAS /* + LI_NO_MSGCONV */;
         }
      }

      } /** ENDOF: IF OK RM 1 **/

      /*** SAVING ORIGINAL DATA IN CASE A REPORT WITH DATA IS REQUIRED ***/
      /*******************************************************************/
      /* The data is saved on the FIRST (or only) gateway processing the */
      /* message. We ensure this with a specific flag:                   */
      /* CSI_QOS_DATA_SAVED of the QoSApplied field of the QoS. If this  */
      /* flag is unset, data has not yet been saved so we are on the 1st */
      /* gateway.                                                        */
      /* Moreover, the Reply-To-Queue specified in the QoS must          */
      /* be located on this very gateway or on a platform depending on   */
      /* this gateway. If one of these conditions is not fulfilled, the  */
      /* report will be created WITHOUT DATA => the ReportRequest field  */
      /* of the QoS of the message is MODIFIED to take this into account */
      /*******************************************************************/

      if ( (LI_ERR_NONE    == status) &&
           (CSIMQMT_REPORT != li_MsgContext_P->MsgDesc_P->MsgType) )
      {
         /* If the message has no data, reports with data are irrelevant, */
         /* thus we toggle off the data bits with a special filter.       */

         if ('\0' != li_MsgContext_P->QoS_P->ReplyToQ[0])
            li_ResolveReplyQ(TRUE, li_MsgContext_P->QoS_P->ReplyToQ);

         if (msgDataDesc_P->DataLen == 0L)
            li_MsgContext_P->QoS_P->ReportRequest &= ~CSIMQRO_DATA_MASK;

         if ( (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DATA_MASK &
               (CSIMQRO_COA_WITH_FULL_DATA|CSIMQRO_COD_WITH_FULL_DATA))
              != 0x00000000L )
         {  /* A COA and/or COD with data is required ... */

            if ( (li_MsgContext_P->QoS_P->AppliedQoS & CSI_QOS_DATA_SAVED)
                 == 0x00000000L )
            {  /* No data has been saved: we are on the 1st gateway. */

               /* Check the Reply To Q */
               
               if ( CSITRUE == li_IsQueueLocal(li_MsgContext_P
                                              ,li_MsgContext_P->QoS_P->ReplyToQ
                                              )
                  )
               {  /* The Reply To Q is OK, going on */

                  ret = li_SaveOriginalData(li_MsgContext_P->TraceInfo_P->UniqueId
                                           ,li_MsgContext_P->QoS_P->ReportRequest
                                           ,(CSIDD*) msgDataDesc_P
                                           ,returnCode_P
                                           ,reasonCode_P
                                           );
                  if (CSISUCCESS == ret)
                     li_MsgContext_P->QoS_P->AppliedQoS |= CSI_QOS_DATA_SAVED;
                  else
                     status = *reasonCode_P;
               }
               else
               {  /* Non-local Reply To Q, report will be without data... */

                  EVR_ERRORLOG(EVR_USERERROR, (LI_MSG_BAD_REPLY_TO_Q,
                               li_MsgContext_P->TraceInfo_P->UniqueId,
                               li_MsgContext_P->QoS_P->ReplyToQ));

/* OR4280
                  li_MsgContext_P->QoS_P->ReportRequest &= ~CSIMQRO_DATA_MASK;
*/
               }
            }
         }
      }

      /*** CALLING THE OPERATION DISPATCHER ***/
      /****************************************/

      if (LI_ERR_NONE == status)
      {
         CSIPDU     *oldmsgDataDesc_P = NULL;

         oldmsgDataDesc_P = msgDataDesc_P;
         ret = li_OperationDispatcher(li_ProcContext_P
                                     ,opFlags
                                     ,li_MsgContext_P
                                     ,&msgDataDesc_P
                                     ,returnCode_P
                                     ,reasonCode_P
                                     );
         if (CSIFAILED==ret)
         {
            status = *reasonCode_P;
         }
         
         if (oldmsgDataDesc_P != msgDataDesc_P && oldmsgDataDesc_P != NULL)
         {
           /* if this is the original LI, the length of the message coming */
           /* from the AP that needs to be audited is in fact the message  */
           /* before any conversion takes place.                           */
           if (isOrigLI)
               lMsgLengthBeforeConversion = oldmsgDataDesc_P->DataLength;

           /* because data conversion can occurs, we must free the old csidd if a 
              new one is returned */

           ret = gcn_FreeDataDesc(&oldmsgDataDesc_P, returnCode_P, reasonCode_P);
           if (CSIFAILED==ret)
           {
             EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
              "gcn_FreeDataDesc",(*reasonCode_P)));
             if (status == LI_ERR_NONE)
             {
              status        = LI_ERR_FREE; /* only return first occured error */
             }
           }
         }
      }

      /*-- INPUT AUDIT TRACE --*/
      /*-----------------------*/

      if (isOrigLI == CSITRUE && get_ok == CSITRUE
         )
      {
         CSILONG  typeM      = 0L;
         CSICHAR8 toGwName   = {""};
         
         typeM = li_MsgContext_P->MsgDesc_P->MsgType;

         if (CSI_REMOTE_EVENT == li_MsgContext_P->EventType)
            CDX_extractNameFromDN(li_MsgContext_P->DestGateway, toGwName);
     
         /* OR5185 : Modify SubmitTimeSec and SubmitTimeUSec in the TraceInfo */
         /* for the reports, so that the submissionTime is properly set       */
         if (CSIMQMT_REPORT == li_MsgContext_P->MsgDesc_P->MsgType)
         {
            struct tm sInputmktime;
	    struct timeval tim;

	    gettimeofday(&tim, NULL);

            sscanf(li_MsgContext_P->MsgDesc_P->PutDate, "%4d%2d%2d"
                  ,&(sInputmktime.tm_year)
                  ,&(sInputmktime.tm_mon)
                  ,&(sInputmktime.tm_mday)
                  );
            sscanf(li_MsgContext_P->MsgDesc_P->PutTime, "%2d%2d%2d"
                  ,&(sInputmktime.tm_hour)
                  ,&(sInputmktime.tm_min)
                  ,&(sInputmktime.tm_sec)
                  );
            sInputmktime.tm_year -= 1900;
            sInputmktime.tm_mon  -= 1;

            li_MsgContext_P->TraceInfo_P->SubmitTimeSec  = tim.tv_sec;
            li_MsgContext_P->TraceInfo_P->SubmitTimeUSec = tim.tv_usec;
         }

         /* Generating "I" trace */
         /* The "I" trace is not more generated just after the CSI_mq_get. */
         /* This will allow to output the "I" trace only when the MQCMIT   */
         /* will take place. In case of a MQBACK (rollback) no audit will  */
         /* be printed.                                                    */

          lSubmissionTime_I = li_MsgContext_P->TraceInfo_P->SubmitTimeSec;
          lSubmissionTimeU_I = li_MsgContext_P->TraceInfo_P->SubmitTimeUSec;
          memcpy(szPeerGateway_I, toGwName, sizeof(szPeerGateway_I));
          memcpy(szUniqueId_I, li_MsgContext_P->TraceInfo_P->UniqueId, sizeof(szUniqueId_I));
          memcpy(szMsgTypeId_I, li_MsgContext_P->MsgTypeId, sizeof(szMsgTypeId_I));
          lMsgLength_I = (-1L == lMsgLengthBeforeConversion)
                             ? (CSILONG) msgDataDesc_P->DataLength
                             : lMsgLengthBeforeConversion;
          lReasonCode_I = status;
          memcpy(szUserName_I, li_MsgContext_P->TraceInfo_P->OrgUserName, sizeof(szUserName_I));
          memcpy(szAppName_I, li_MsgContext_P->TraceInfo_P->OrgAppliName, sizeof(szAppName_I));
          memcpy(szQueueName_I, li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName, sizeof(szQueueName_I));

          lMsgType_I = typeM;
          memcpy(szMsgId_I, li_MsgContext_P->MsgDesc_P->MsgId, sizeof(szMsgId_I));
          memcpy(szCorrelId_I, li_MsgContext_P->MsgDesc_P->CorrelId, sizeof(szCorrelId_I));
          lFeedback_I = li_MsgContext_P->MsgDesc_P->Feedback;
          memcpy(szReplyToQueue_I, li_MsgContext_P->QoS_P->ReplyToQ, sizeof(szReplyToQueue_I));
          lReportRequest_I = li_MsgContext_P->QoS_P->ReportRequest;
      }

      /*** PUTTING THE MESSAGE IN THE NEXT HOP QUEUE ***/
      /*************************************************/

      /** Building the new Trace Information **/

      if (LI_ERR_NONE == status)
      {
         /* For Async messages we will always send V1 of the Traceinfo structure */
         li_MsgContext_P->TraceInfo_P->Version = CSIIITFHDR_VERSION_1;

         ret = gcn_BuildTraceInfo(li_MsgContext_P->EventType,
               li_MsgContext_P->TraceInfo_P,
               NULL,li_MsgContext_P->UserProfileList_P,
               li_MsgContext_P->PrevGatewayList_P, msgDataDesc_P, 
               returnCode_P,reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_BuildTraceInfo",(*reasonCode_P)));
            status        = LI_ERR_INTERNAL;
         } 

         /* Trace info put in message */
         li_MsgContext_P->isTrcInfoInMsg = CSITRUE;

      }

      /** Modification of the automatical report options of the CSIMQMD **/
    
      /* Identify Local or Remote Destination Gateway */

      if (LI_ERR_NONE == status)
      {
         ret = CDX_isLocalGateway(li_MsgContext_P->DestGateway,&toLocal,
                                  returnCode_P, reasonCode_P);
         if (CSIFAILED == ret)
         {
            status = *reasonCode_P;
         }
      }


      /* If the message is not a report: prepare the report generation */
      /* if need be. If it is a report, check whether it is necessary  */
      /* to restore previously saved data (for report with data, of    */
      /* course).                                                      */

      if (LI_ERR_NONE == status)
      {
         if (CSIMQMT_REPORT != li_MsgContext_P->MsgDesc_P->MsgType)
         {  /* Not a report: prepare MQS report generation */

            interface = ( (TRUE == toLocal) ? ROU_LI_TO_EXT
                                            : ROU_LI_TO_XTA );

            ret = rou_SetReportingEOS(li_MsgContext_P->QoS_P,
                  interface, li_MsgContext_P->MsgDesc_P,
                  returnCode_P, reasonCode_P);
            if (CSIFAILED == ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_ROU_ERROR,
               "rou_SetReportEOS",(*reasonCode_P)));
               status = *reasonCode_P;
            }
         }
         else
         {  /* It is a report: is data required ? Must it be restored ? */

            if (   ((li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DATA_MASK) != CSIMQRO_NONE) 
                && (TRUE == toLocal) 
                /* OR 6826 : to call li_RestoreOriginalData(),                  */
                /* a CSIMQRO_COA_WITH_FULL_DATA or a CSIMQRO_COD_WITH_FULL_DATA */
                /* must have been called.                                       */
                && (   ((li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_COA_WITH_FULL_DATA) == CSIMQRO_COA_WITH_FULL_DATA)
                    || ((li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_COD_WITH_FULL_DATA) == CSIMQRO_COD_WITH_FULL_DATA)
                   )
               )
            {  /* Data is required, it must be restored (to Local) */

               ret = li_RestoreOriginalData(li_MsgContext_P->TraceInfo_P->UniqueId
                                           ,li_MsgContext_P->QoS_P->ReportRequest
                                           ,li_MsgContext_P->MsgDesc_P->Feedback
                                           ,msgDataDesc_P
                                           ,returnCode_P
                                           ,reasonCode_P
                                           );
               if (CSIFAILED == ret)
               {
                  status = *reasonCode_P;
               }
               else
               {
                  /* Following check determinates if the report was required */
                  /* or should be dismissed (meaning the Put is cancelled).  */

                  cancelPutReport = li_ShouldReportBeDismissed(
                     li_MsgContext_P->QoS_P->ReportRequest,
                     li_MsgContext_P->MsgDesc_P->Feedback);
               }
            }
         }    
      }
    
      /** Modification of the queue address **/

      if (LI_ERR_NONE == status)
      {
         strncpy(objDesc_S.ObjectName
                ,li_MsgContext_P->NextHopName
                ,sizeof(CSICHAR48)
                );
      }

      /** PUTting with CSI_mq_put1 **/

      if ( (LI_ERR_NONE == status) && (cancelPutReport == CSIFALSE) )
      {
         ret = CSI_mq_put1(li_ProcContext_P->QMgrHandle
                          ,&objDesc_S
                          ,li_MsgContext_P->MsgDesc_P
                          ,&putMsgOpt_S
                          ,(CSIDD *)msgDataDesc_P
                          ,li_MsgContext_P->QoS_P
                          ,returnCode_P
                          ,reasonCode_P
                          );

         /*--- INPUT AUDIT TRACE ---*/
         /*-------------------------*/
         if (CSISUCCESS == ret)
         {
            if ((isOrigLI == CSITRUE) && (get_ok == CSITRUE))
            {
                li_PrepAudit(lSubmissionTime_I, lSubmissionTimeU_I, tt2);
                li_PrepAudit(0L, 0L, tt3);

                EVR_ASYNC_AUDIT_IN_FROM_AP(lSubmissionTime_I
                                      ,szPeerGateway_I[0] == '\0' ? getenv(CDX_C_ENV_GATEWAY)
                                                                  : szPeerGateway_I          
                                      ,szUniqueId_I
                                      ,szMsgTypeId_I
                                      ,lMsgLength_I 
                                      ,lReasonCode_I == LI_ERR_NONE ? CSISUCCESS : CSIFAILED
                                      ,lReasonCode_I
                                      ,szUserName_I 
                                      ,szAppName_I 
                                      ,szQueueName_I
                                      ,EVR_FORMATASYNCTYPE(lMsgType_I, lFeedback_I)
                                      ,szMsgId_I
                                      ,szCorrelId_I
                                      ,lFeedback_I
                                      ,szReplyToQueue_I[0] == '@' ? "" : szReplyToQueue_I
                                      ,lReportRequest_I
				      ,tt2
				      ,tt3
                                      ,CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode)
                                      );
            }
            EVR_TRACELOG (EVR_DEBUG, (LI_TRC_PUT_IN_Q
                                     ,li_MsgContext_P->TraceInfo_P->UniqueId
                                     ,li_MsgContext_P->NextHopName
                                     )
                         );      
         }

         /*--- OUTPUT AUDIT TRACE ---*/
         /*--------------------------*/
         if (
                li_MsgContext_P->EventType == CSI_LOCAL_EVENT
             && CSISUCCESS == ret /* Generating an "O" trace only if CSI_mq_put1 was OK */
            )
         {
            CSILONG  typeM      = 0;
            CSICHAR8 fromGwName = {""};
	    
            struct timeval tim;
            gettimeofday(&tim, NULL);

            li_PrepAudit(li_MsgContext_P->TraceInfo_P->SubmitTimeSec
                         ,li_MsgContext_P->TraceInfo_P->SubmitTimeUSec
                         ,tt2            
                        );     

            li_PrepAudit(tim.tv_sec, tim.tv_usec, tt3);

            typeM = li_MsgContext_P->MsgDesc_P->MsgType;

            CDX_extractNameFromDN(li_MsgContext_P->OrigGateway, fromGwName);

            EVR_ASYNC_AUDIT_OUT_TO_AP(li_MsgContext_P->TraceInfo_P->SubmitTimeSec,
                                     (0 == strcmp("", fromGwName) ? getenv(CDX_C_ENV_GATEWAY) : fromGwName),
                                     li_MsgContext_P->TraceInfo_P->UniqueId,
                                     li_MsgContext_P->MsgTypeId,
                                     msgDataDesc_P->DataLength,
                                     *returnCode_P,
                                     *reasonCode_P,
                                     li_MsgContext_P->TraceInfo_P->OrgUserName,
                                     li_MsgContext_P->TraceInfo_P->OrgAppliName,
                                     objDesc_S.ObjectName,
                                     EVR_FORMATASYNCTYPE(typeM,li_MsgContext_P->MsgDesc_P->Feedback),
                                     li_MsgContext_P->MsgDesc_P->MsgId,
                                     li_MsgContext_P->MsgDesc_P->CorrelId,
                                     li_MsgContext_P->MsgDesc_P->Feedback,
                                     (li_MsgContext_P->QoS_P->ReplyToQ[0] == '@' ? "" : li_MsgContext_P->QoS_P->ReplyToQ),
                                     li_MsgContext_P->QoS_P->ReportRequest,
				     tt2,
				     tt3,
                                     CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode)
				     );

            /*--- End Audit Trace ---*/
         }

         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG (EVR_MAJOR, (LI_MSG_NEXTQ_PUT_ERROR
                                     ,li_MsgContext_P->NextHopName
                                     ,*reasonCode_P
                                     )
                         );
            status = *reasonCode_P;
         }
      }

      /*** PUTTING THE MESSAGE IN A DEAD LETTER QUEUE IF NECESSARY ***/
      /***************************************************************/
      /* This step is shunned if the CSI_mq_get() failed - for any reason */
      /* It will be performed only if an error occurred, but AFTER the    */
      /* CSI_mq_get(), so we are supposed to have a message...            */

      if ( (LI_ERR_NONE != status) && (CSITRUE == get_ok) )
      {
         if (0 == (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_DISCARD_MSG))
         {
            CSILONG Report = CSIMQRO_NONE, Expiry = 0;
            CSILONG Qmode;

            /** Fetch address of the CSI Dead Letter Queue through CDX **/
            /* get DLQ for the RunMode of msg - LVR OR7172 */
            switch (msgDataDesc_P->Flags & CSI_DD_FLAGS_MODE_SET)
               {
               case CSI_DD_FLAGS_PRODUCTION_MSG:
                  Qmode = CSI_DLP;
                  break;

               case CSI_DD_FLAGS_BACKUP_MSG:
                  Qmode = CSI_DLB;
                  break;

               case CSI_DD_FLAGS_TEST_MSG:
                  Qmode = CSI_DLT;
                  break;

               case CSI_DD_FLAGS_MAINTENANCE_MSG:
                  Qmode = CSI_DLM;
                  break;

               default: /* no flag set */
                  Qmode = CSI_DDL;
               }

            ret = CDX_getCSINameOfCoT(NULL
                                     ,Qmode
                                     ,li_MsgContext_P->QoS_P->CoT
                                     ,&deadLetterQName
                                     ,returnCode_P
                                     ,reasonCode_P
                                     );
            if (CSISUCCESS == ret)
            {
               /** Copy the Dead-Letter queue name into the Object Desc. **/

               strncpy(objDesc_S.ObjectName, deadLetterQName, sizeof(CSICHAR48));

               /* Saving report and expiry for report generation */
               Report = li_MsgContext_P->MsgDesc_P->Report;
               Expiry = li_MsgContext_P->MsgDesc_P->Expiry;

               /* Reset report and expiry prior to put msg in dlq */
               li_MsgContext_P->MsgDesc_P->Report = CSIMQRO_NONE;
               li_MsgContext_P->MsgDesc_P->Expiry = CSIMQEI_UNLIMITED;

               /* START OR5345 */
               /* Put Trace info if it doesn't exist */
               if (CSITRUE != li_MsgContext_P->isTrcInfoInMsg)
               {
                  ret = gcn_BuildTraceInfo(li_MsgContext_P->EventType,
                                           li_MsgContext_P->TraceInfo_P,
                                           NULL,
                                           li_MsgContext_P->UserProfileList_P,
                                           li_MsgContext_P->PrevGatewayList_P,
                                           msgDataDesc_P,
                                           returnCode_P, reasonCode_P) ;
                  if (CSIFAILED == ret)
                  {
                     EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_GCN_ERROR,
                                                      "gcn_BuildTraceInfo",
                                                      (*reasonCode_P))) ;
                     status = *reasonCode_P;
                  }
                  else
                     /* Trace info put in message */
                     li_MsgContext_P->isTrcInfoInMsg = CSITRUE ;
               }
               /* END OR5345 */
            }
            else
            {
               EVR_ERRORLOG (EVR_UNDETERMINATE, (LI_MSG_CDX_ERROR
                                                ,"CDX_getCSINameOfCoT"
                                                ,*reasonCode_P
                                                )
                            );
               status = *reasonCode_P;
            }

            if (ret == CSISUCCESS)
            {
               /** mq_put1 the message in the found Dead Letter Q **/

               ret = CSI_mq_put1(li_ProcContext_P->QMgrHandle
                                ,&objDesc_S
                                ,li_MsgContext_P->MsgDesc_P
                                ,&putMsgOpt_S
                                ,(CSIDD *)msgDataDesc_P
                                ,li_MsgContext_P->QoS_P
                                ,returnCode_P
                                ,reasonCode_P
                                );

               /* Restoring report and expiry for report generation */
               li_MsgContext_P->MsgDesc_P->Report = Report;
               li_MsgContext_P->MsgDesc_P->Expiry = Expiry;

               if (CSIFAILED == ret)
               {
                  EVR_ERRORLOG (EVR_MAJOR,(LI_MSG_NEXTQ_PUT_ERROR
                                          ,deadLetterQName
                                          ,*reasonCode_P
                                          )
                               );
                  status = *reasonCode_P;
               }
            }

            if (ret == CSISUCCESS)
            {
               CSILONG  typeM      = 0L;
               CSICHAR8 fromGwName = {""};

               EVR_ERRORLOG(EVR_MAJOR, (LI_MSG_DEAD_QUEUE_PUT, status));     
               dead_queue = CSITRUE;

               /*-- AUDIT & STATS TRACING --*/
               /*---------------------------*/

               typeM = li_MsgContext_P->MsgDesc_P->MsgType;

               CDX_extractNameFromDN(li_MsgContext_P->OrigGateway, fromGwName);

               /* AUDIT: Generating traces */

               if (isOrigLI == CSITRUE)
               {

                   li_PrepAudit(lSubmissionTime_I, lSubmissionTimeU_I, tt2);
                   li_PrepAudit(0L, 0L, tt3);

                   EVR_ASYNC_AUDIT_IN_FROM_AP(lSubmissionTime_I
                                             ,szPeerGateway_I[0] == '\0' ? getenv(CDX_C_ENV_GATEWAY)
                                                                         : szPeerGateway_I
                                             ,szUniqueId_I
                                             ,szMsgTypeId_I
                                             ,lMsgLength_I
                                             ,lReasonCode_I == LI_ERR_NONE ? CSISUCCESS : CSIFAILED
                                             ,lReasonCode_I
                                             ,szUserName_I
                                             ,szAppName_I
                                             ,szQueueName_I
                                             ,EVR_FORMATASYNCTYPE(lMsgType_I, lFeedback_I)
                                             ,szMsgId_I
                                             ,szCorrelId_I
                                             ,lFeedback_I
                                             ,szReplyToQueue_I[0] == '@' ? "" : szReplyToQueue_I
                                             ,lReportRequest_I
                                             ,tt2             
                                             ,tt3
                                             ,CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode)
                                             );
               }

               struct timeval tim;
               gettimeofday(&tim, NULL);

               li_PrepAudit(li_MsgContext_P->TraceInfo_P->SubmitTimeSec
                            ,li_MsgContext_P->TraceInfo_P->SubmitTimeUSec
                            ,tt2
                           );
               li_PrepAudit(tim.tv_sec, tim.tv_usec, tt3);


               EVR_ASYNC_AUDIT_OUT_TO_AP(li_MsgContext_P->TraceInfo_P->SubmitTimeSec,
                                   (0 == strcmp("", fromGwName) ? getenv(CDX_C_ENV_GATEWAY) : fromGwName),
                                   li_MsgContext_P->TraceInfo_P->UniqueId,
                                   li_MsgContext_P->MsgTypeId,
                                   msgDataDesc_P->DataLength,
                                   CSIFAILED,
                                   status,
                                   li_MsgContext_P->TraceInfo_P->OrgUserName,
                                   li_MsgContext_P->TraceInfo_P->OrgAppliName,
                                   objDesc_S.ObjectName,
                                   EVR_FORMATASYNCTYPE(typeM,li_MsgContext_P->MsgDesc_P->Feedback),
                                   li_MsgContext_P->MsgDesc_P->MsgId,
                                   li_MsgContext_P->MsgDesc_P->CorrelId,
                                   li_MsgContext_P->MsgDesc_P->Feedback,
                                   (li_MsgContext_P->QoS_P->ReplyToQ[0] == '@' ? "" : li_MsgContext_P->QoS_P->ReplyToQ),
                                   li_MsgContext_P->QoS_P->ReportRequest,
                                   tt2,
                                   tt3,
                                   CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode));
 
               /*--- End Audit Trace ---*/
            }
            else
            {
                /* an error occured after the messages was retrieved from the */
                /* queue, so the MQ-Series transaction must be aborted and    */
                /* the audit lines must not be written.                       */
                rollbackTransaction = CSITRUE;
            }
         }      
         else
         {
            /* The CSIMQRO_DISCARD_MSG options was set in the ReportRequest */

            CSILONG  typeM      = 0L;
            CSICHAR8 fromGwName = {""};

            EVR_ERRORLOG(EVR_MAJOR, (LI_MSG_DISCARDED, status));

            /*-- AUDIT & STATS TRACING --*/
            /*---------------------------*/

            typeM = li_MsgContext_P->MsgDesc_P->MsgType;

            CDX_extractNameFromDN(li_MsgContext_P->OrigGateway, fromGwName);

            /* AUDIT: Generating traces */

            if (isOrigLI == CSITRUE)
            {

                li_PrepAudit(lSubmissionTime_I, lSubmissionTimeU_I, tt2);
                li_PrepAudit(0L, 0L, tt3);

                EVR_ASYNC_AUDIT_IN_FROM_AP(lSubmissionTime_I
                                          ,szPeerGateway_I[0] == '\0' ? getenv(CDX_C_ENV_GATEWAY)
                                                                      : szPeerGateway_I
                                          ,szUniqueId_I
                                          ,szMsgTypeId_I
                                          ,lMsgLength_I
                                          ,lReasonCode_I == LI_ERR_NONE ? CSISUCCESS : CSIFAILED
                                          ,lReasonCode_I
                                          ,szUserName_I
                                          ,szAppName_I
                                          ,szQueueName_I
                                          ,EVR_FORMATASYNCTYPE(lMsgType_I, lFeedback_I)
                                          ,szMsgId_I
                                          ,szCorrelId_I
                                          ,lFeedback_I
                                          ,szReplyToQueue_I[0] == '@' ? "" : szReplyToQueue_I
                                          ,lReportRequest_I
                                          ,tt2             
                                          ,tt3
                                          ,CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode)
                                          );
            }

            struct timeval tim;
            gettimeofday(&tim, NULL);
                                     
            li_PrepAudit(li_MsgContext_P->TraceInfo_P->SubmitTimeSec
                         ,li_MsgContext_P->TraceInfo_P->SubmitTimeUSec
                         ,tt2
                        );
            li_PrepAudit(tim.tv_sec, tim.tv_usec, tt3);

            EVR_ASYNC_AUDIT_OUT_TO_AP(li_MsgContext_P->TraceInfo_P->SubmitTimeSec,
                                (0 == strcmp("", fromGwName) ? getenv(CDX_C_ENV_GATEWAY) : fromGwName),
                                li_MsgContext_P->TraceInfo_P->UniqueId,
                                li_MsgContext_P->MsgTypeId,
                                msgDataDesc_P->DataLength,
                                CSIFAILED,
                                status,
                                li_MsgContext_P->TraceInfo_P->OrgUserName,
                                li_MsgContext_P->TraceInfo_P->OrgAppliName,
                                "MSG_DISCARDED",
                                EVR_FORMATASYNCTYPE(typeM,li_MsgContext_P->MsgDesc_P->Feedback),
                                li_MsgContext_P->MsgDesc_P->MsgId,
                                li_MsgContext_P->MsgDesc_P->CorrelId,
                                li_MsgContext_P->MsgDesc_P->Feedback,
                                (li_MsgContext_P->QoS_P->ReplyToQ[0] == '@' ? "" : li_MsgContext_P->QoS_P->ReplyToQ),
                                li_MsgContext_P->QoS_P->ReportRequest,
                                tt2,
                                tt3,
                                CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode));
            /*--- End Audit Trace ---*/
         }
      }

      /*---- Check if file associated to datadescriptor 
             do not need to be deleted. Keep the file if the message is put in
             the DDL queue. Deletes the file before sending the report as the
             rou_SendReport() function resets the message Data Descriptor. -----*/
             /* File shoudl be deleted only in there is no EXCEP report with data
                required */
      if (   LI_ERR_NONE != status
          && CSITRUE == get_ok 
          && dead_queue == CSIFALSE
           /* data have to be transferred by file to delete a file ! */
          && CSI_DD_FLAGS_FILE_STORAGE(msgDataDesc_P)
          && (CSIMQRO_EXCEPTION_WITH_FULL_DATA != (  li_MsgContext_P->QoS_P->ReportRequest 
                                                   & CSIMQRO_EXCEPTION_WITH_FULL_DATA
                                                  )
             )
         )
      {
         ret = gcn_DeleteFile(msgDataDesc_P, &tmpRet, &tmpReas);
      }

      /*---- Check if file associated to datadescriptor needs to be kept.
             In case of an exception report, if the message is put in the
             DLQ */ 
      if (   LI_ERR_NONE != status
          && CSITRUE == get_ok 
          && dead_queue == CSITRUE 
           /* data have to be transferred by file to create a link on a file ! */
          && CSI_DD_FLAGS_FILE_STORAGE(msgDataDesc_P)
          && (CSIMQRO_EXCEPTION_WITH_FULL_DATA == (  li_MsgContext_P->QoS_P->ReportRequest
                                                   & CSIMQRO_EXCEPTION_WITH_FULL_DATA
                                                  )
             )
         )   
      {
         strcpy(szReportFileName, msgDataDesc_P -> FileName);
         strcat(szReportFileName, "=");

         /* Create a hard link on the file */
         ret = link(msgDataDesc_P -> FileName, szReportFileName);

         if (ret == -1) status = LI_ERR_INTERNAL;
      }
      else
      {
         strcpy(szReportFileName, msgDataDesc_P -> FileName);
      }

      /* All the operations concerning CSI_mq_get & CSI_mq_put1 are now  */
      /* finished so the MQ-Series transaction is committed/rollabacked. */

      if (CSITRUE == rollbackTransaction)
      {
         MQBACK(li_ProcContext_P->QMgrHandle, returnCode_P, reasonCode_P);
      }
      else
      {
         MQCMIT(li_ProcContext_P->QMgrHandle, returnCode_P, reasonCode_P);
      }

      /*** SENDING A REPORT IN CASE OF FAILURE ***/
      /*******************************************/
      /* This step is shunned if the CSI_mq_get() failed - for any reason */
      /* It will be performed only if an error occurred, but AFTER the    */
      /* CSI_mq_get(), so we are supposed to have a message...            */
      /* If status equals CSI_ERR_SWITCHED, this means that the message   */
      /* was in degraded mode, and that the switch-back function failed   */
      /* and sent the message back in the degraded queue.                 */
      /* If rollbackTransaction is true, this means that a MQ-Series      */
      /* rollback (MQBACK) will be taken and that the message will not be */
      /* deleted from the queue, so no action is taken.                   */

      if (   LI_ERR_NONE      != status
          && CSI_ERR_SWITCHED != status
          && CSITRUE          == get_ok
          && CSIFALSE         == rollbackTransaction
         )
      {

         /* We've to generate a report. First check if the trace info is still
            in message, and if not, put it back */
         if( ! li_MsgContext_P->isTrcInfoInMsg )
         {
            /* both context's fields have to be set to zero because */
            /* null pointers are passed to the gcn_BuildTraceInfo.  */
            li_MsgContext_P->TraceInfo_P->NbOfPrevGateways = 0;
            li_MsgContext_P->TraceInfo_P->NbOfOrgUserProfiles = 0;

            ret = gcn_BuildTraceInfo(CSI_LOCAL_EVENT,
                                     li_MsgContext_P->TraceInfo_P, 
                                     NULL, /* triggerQName     */
                                     NULL, /* orgUserProfBuf_P */
                                     NULL, /*prevGatewayBuf_P  */
                                     msgDataDesc_P, returnCode_P,reasonCode_P);

            if (CSIFAILED==ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
               "gcn_BuildTraceInfo",(*reasonCode_P)));
               status        = LI_ERR_INTERNAL;
            }
         }

         if(CSISUCCESS  == *returnCode_P)
         {
            strcpy(msgDataDesc_P->FileName, szReportFileName);

            if ( (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_EXCEPTION_WITH_FULL_DATA)
                       == CSIMQRO_EXCEPTION_WITH_FULL_DATA)
            {
               ret = rou_SendReport(NULL,li_ProcContext_P->QMgrHandle,
                     ROU_EXCEPTION_WITH_FULL_DATA, (status + CSIMQFB_CCN_FIRST),
                     li_MsgContext_P->TraceInfo_P->UniqueId,
                     li_MsgContext_P->QoS_P, li_MsgContext_P->MsgDesc_P,
                     msgDataDesc_P, returnCode_P, reasonCode_P);
            }
            else if (li_MsgContext_P->QoS_P->ReportRequest & CSIMQRO_EXCEPTION)
            {
               ret = rou_SendReport(NULL,li_ProcContext_P->QMgrHandle,
                     ROU_EXCEPTION, (status + CSIMQFB_CCN_FIRST),
                     li_MsgContext_P->TraceInfo_P->UniqueId,
                     li_MsgContext_P->QoS_P, li_MsgContext_P->MsgDesc_P,
                     msgDataDesc_P, returnCode_P, reasonCode_P);
            }
         }
      }

      /*** RELEASING THE DATA DESCRIPTOR ***/
      /*************************************/

      ret = gcn_FreeDataDesc(&msgDataDesc_P, returnCode_P, reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
         "gcn_FreeDataDesc",(*reasonCode_P)));
         status        = LI_ERR_FREE;
      }

      /*** END OF THE RETRIEVING LOOP ***/
      /**********************************/
      /*****************************************************************/
      /* WARNING: (status != LI_ERR_NONE) is the exit condition of the */
      /* ======== loop. For us the exit conditions are:                */
      /*          - either the CSI_mq_get() failed or                  */
      /*          - the CSI_mq_get() was not performed.                */
      /*          - a MQ-Series rollback (MQBACK) was taken            */
      /*                                                               */
      /*          All other errors do not cause the end of the loop,   */
      /*          so status will be reset to LI_ERR_NONE according to  */
      /*          the values of boolean get_done and get_ok.           */
      /*****************************************************************/

      if (   CSITRUE  == get_done
          && CSITRUE  == get_ok
          && CSIFALSE == rollbackTransaction
         )
      {
         status = LI_ERR_NONE;
      }
      else if (CSITRUE == rollbackTransaction)
      {
         /* the status is set so that a tpreturn with a CSI_TPEXIT */
         /* will be called.                                        */
         status = LI_ERR_ROLLBACK;
      }

   } /** ENDOF: RETRIEVING LOOP **/

   /* 'No message available' is not actually an error, so...: */

   if (MQ_ERR_NO_MSG_AVAILABLE==status) status = LI_ERR_NONE;


   /*** CLOSING THE TRIGGERED QUEUE ***/
   /***********************************/

   c_Options = 0L;           /* No close options for now */

   if (open_ok == CSITRUE)
   {
      /* before closing the queue the trigger type attribute of the queue */
      /* has to be set to EVERY - the reverse operation ( set to FIRST )  */
      /* is done by the trigger monitor                                   */

      attrSelectorSet[ALI_MQSET_TRIGTYPE_ATTR_INDEX] = MQIA_TRIGGER_TYPE;
      attrValuesSet[ALI_MQSET_TRIGTYPE_ATTR_INDEX] = MQTT_EVERY;
      MQSET (li_ProcContext_P->QMgrHandle, objHandle ,
             ALI_MQSET_REQUESTED_ATTR_MXNUM, attrSelectorSet,
             ALI_MQSET_REQUESTED_ATTR_MXNUM, attrValuesSet,
             0, NULL, &mqResultCode, &mqReasonCode);

      if (mqResultCode == MQCC_FAILED)
      {
         /* Failed to set on target queue. */
         EVR_ERRORLOG (EVR_MINOR, (LI_MSG_MQ_SET_FAILED,
                                   li_MsgContext_P->TriggerQName, mqReasonCode));
      }


      ret = CSI_mq_close(li_ProcContext_P->QMgrHandle, &objHandle, c_Options,
                         returnCode_P, reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
         "CSI_mq_close",(*reasonCode_P)));
         status = *reasonCode_P;
      }
   }

   /*** OR4237: DISCONNECTING FROM THE LOCAL QUEUE MANAGER ***/
   /**********************************************************/

   if(CSIMQHC_UNUSABLE_HCONN != li_ProcContext_P->QMgrHandle)
   {
      ret = CSI_mq_disc(&(li_ProcContext_P->QMgrHandle),returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
         "CSI_mq_disc",*reasonCode_P));
         status = *reasonCode_P;
      }
   }
   
   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_AsyncProc",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}


/******************************************************/
/*** SAVE ORIGINAL DATA : CASE OF REPORTS WITH DATA ***/
/******************************************************/
/*
* @LOCALFUN li_SaveOriginalData
*
* @DESCRIPTION Saves original data of a message (for reports with data)
*
* @ABSTRACT
*
* This function stores the data of a CSIDD into a file. If this data
* already is in a file, a copy will be made. The name of the file is
* based on the Unique Id of the message (see <@LOCALFUN li_BuildFileName>).
* Those temporary files are stored in a subdirectory of the general
* temporary storage area of the gateway (see <@MACRODEF LI_WKPATH>).
*
* The main purpose of this function is to save the original data in
* case a report with data is required in the message. Thus, this
* data is needed in order to build the report. Keep in mind that the data
* of the message itself might be modified (data presentation for
* instance), and it is thus necessary to keep a copy of the original
* before any transformation. This function is called by 
* <@LOCALFUN li_AsyncProc> just after the retrieval of the message.
*
* @RETURNVALUE CSILONG
* Possible returnvalues: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* IN   CSICHAR48  uniqueId
*    The Unique Identifier of the message.
*
* IN   CSILONG    reportOptions
*    The report options initially required.
*
* INOUT CSIDD *dataDesc_P
*       A pointer to a CSIDD structure. In output the CSIDD will
*       contain the filename of the copy.
*
* OUT   CSILONG *returnCode_P
*
* OUT   CSILONG *reasonCode_P
*
* @SEEALSO
* <@LOCALFUN li_AsyncProc>
*/
CSILONG   li_SaveOriginalData(
IN        CSICHAR48    uniqueId,
IN        CSILONG      reportOptions,
INOUT     CSIDD       *dataDesc_P,
OUT       CSILONG     *returnCode_P,
OUT       CSILONG     *reasonCode_P
)
{
   /** LOCAL VARIABLES DECLARATION **/
   /*********************************/

   CSILONG    status      = LI_ERR_NONE;
   CSILONG    ret         = CSISUCCESS;

   CSILONG    reportFlags = 0L;

   CSICHAR256 tmpFileName = {""};
   CSICHAR*   tmpName_P   = NULL;
   CSILONG    fileDesc    = 0L;
   CSIULONG   write_len   = 0L;

   TRACE_INT_LOW_FUNCTION_ENTRY("li_SaveOriginalData");

   /*****************************************************/
   /* Filter report options to keep only relevant bits. */
   /* By 'relevant bits' we mean bits associated with   */
   /* reports with data. Note that in the MQRO, each    */
   /* one of the four possible report types (COA, COD,  */
   /* Expiration and Exception) is associated with 3    */
   /* bits : the first one indicates that this type of  */
   /* report is required, the second one adds that data */
   /* is required (100 first bytes), the third one adds */
   /* that all data is required.                        */
   /* CSIMQRO_MASK is the combination of all second and */
   /* third bits of the four report types. When this    */
   /* filter is applied, we thus only keep the bits as- */
   /* -sociated with report with data (partial or total)*/
   /*****************************************************/

   reportFlags = (reportOptions & CSIMQRO_DATA_MASK);

   /** CREATION OF THE FILE **/
   /**************************/

   tmpName_P = li_BuildFileName(uniqueId, tmpFileName);
   if (NULL == tmpName_P)
   {
      status = LI_ERR_NO_FILE_NAME;
   }

   /** COPY ALL DATA INTO THE FILE **/
   /*********************************/

   if (LI_ERR_NONE == status)
   {
      if ( CSI_DD_FLAGS_MEMORY_STORAGE(dataDesc_P) )
      {  /* Transferring data from memory to the file */
         ret = OS_storage_open(tmpFileName, (OS_O_CREAT | OS_O_WRONLY),
               &fileDesc, returnCode_P, reasonCode_P);
         if (CSIFAILED == ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_OS_ERROR, 
                         "OS_storage_open", (*reasonCode_P)));
           status = *reasonCode_P;
         }
         else
         {
            write_len = dataDesc_P->DataLen;
            ret = OS_storage_write(fileDesc, &write_len, dataDesc_P->Data,
                                   returnCode_P, reasonCode_P);
            if (CSIFAILED == ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_OS_ERROR,
                            "OS_storage_write", (*reasonCode_P)));

               status = *reasonCode_P;
            }
         }
      }
      else
      {  /* Copying the data file */
         ret = gcn_CopyFile(dataDesc_P->FileName, tmpFileName, 0L, 0L,
               dataDesc_P->DataLength, returnCode_P, reasonCode_P);
         if (CSIFAILED == ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_GCN_ERROR,
                         "gcn_CopyFile", (*reasonCode_P)));

            status = *reasonCode_P;
         }
         else
         { /* Re-opens the file with "append" options */
            ret = OS_storage_open(tmpFileName, OS_O_WRONLY | OS_O_APPEND,
                  &fileDesc, returnCode_P, reasonCode_P);
            if (CSIFAILED == ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_OS_ERROR,
                            "OS_storage_open", (*reasonCode_P)));
            }
         }
      }

      if (LI_ERR_NONE == status)
      { /*************************************************/
        /* Adding a set of flags at the end of the file. */
        /* This set indicates which reports have yet to  */
        /* be generated with data.                       */
        /*************************************************/
         write_len = sizeof(CSILONG);
 
         ret = OS_storage_write(fileDesc, &write_len, &reportFlags,
                                returnCode_P, reasonCode_P);
         if (CSIFAILED == ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_GCN_ERROR,
                         "gcn_CopyFile", (*reasonCode_P)));

            status = *reasonCode_P;
         }
         else
         {
            EVR_TRACELOG(EVR_DEBUG, (LI_TRC_ORIGINAL_DATA_SAVED,
                         dataDesc_P->DataLength, tmpFileName));
         }
      }
   }

   /** CLOSE THE FILE **/
   /********************/

   if (fileDesc > 0)
   {
      ret = OS_storage_close(fileDesc, OS_C_CLOSE, returnCode_P, reasonCode_P);
   }

   /** THE END **/
   /*************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_INT_LOW_FUNCTION_EXIT("li_SaveOriginalData",
                               *returnCode_P, *reasonCode_P);
   return(*returnCode_P);
}


/*******************************************************/
/*** RESTORE ORIGINAL DATA: CASE OF REPORT WITH DATA ***/
/*******************************************************/
/*
* @LOCALFUN li_RestoreOriginalData
*
* @DESCRIPTION Restores or deletes the original data for a report.
*
* @ABSTRACT
* This function is symmetrical to <@LOCALFUN li_SaveOriginalData>;
* it either inserts in the report or deletes the data which was saved by this
* function, according to the required set of reports and amount of
* data. It is also called by <@LOCALFUN li_AsyncProc>.
*
* @RETURNVALUE CSILONG
*
* @PARAMETER
* IN     CSICHAR48 uniqueId
*        The Unique Id associated with the message (field of the
*        CSIIITFHDR structure, trace info).
*
* IN     CSILONG reportOptions
*        The field ReportRequest of the QOS of the report.
*
* IN     CSILONG reportFeedback
*        The field Feedback of the MQMD of the report.
*
* INOUT  CSIDD *dataDesc_P
*        CSIDD of the report. The fields FileName and DataLength
*        are modified.
*
* OUT    CSILONG *returnCode_P
*
* OUT    CSILONG *reasonCode_P
*
* @SEEALSO
* <@LOCALFUN li_SaveOriginalData>
*/
CSILONG   li_RestoreOriginalData(
IN        CSICHAR48   uniqueId,
IN        CSILONG     reportOptions,
IN        CSILONG     reportFeedback,
INOUT     CSIPDU     *dataDesc_P,
OUT       CSILONG    *returnCode_P,
OUT       CSILONG    *reasonCode_P
)
{
   /** LOCAL VARIABLES DECLARATION **/
   /*********************************/

   CSILONG    status       = LI_ERR_NONE;
   CSILONG    ret          = CSISUCCESS; 

   CSICHAR256 tmpFileName  = {""};
   CSICHAR256 tmpLinkName  = {""};
   CSICHAR   *tmpName_P    = NULL;
   CSILONG    fileDesc     = 0L;

   CSIULONG   length       = 0L;

   CSIBOOLEAN restoreFile  = CSIFALSE;
   CSIBOOLEAN deleteFile   = CSIFALSE;

   CSIBOOLEAN partialData  = CSIFALSE;

   CSILONG    dataOptions  = 0L;
   CSILONG    reportType   = 0L;
   CSILONG    reportFlags  = 0L;

   TRACE_INT_LOW_FUNCTION_ENTRY("li_RestoreOriginalData");

   /* Filter report options to keep only relevant bits */

   dataOptions = (reportOptions & CSIMQRO_DATA_MASK);

   /** FEEDBACK ANALYSIS **/
   /***********************/

   switch(reportFeedback)
   {
      case CSIMQFB_COA:
      if ((dataOptions & CSIMQRO_COA_WITH_DATA)!=0L) 
      {
         reportType  = CSIMQRO_COA_WITH_FULL_DATA;
         restoreFile = CSITRUE;
         if ((reportOptions & CSIMQRO_COA_WITH_FULL_DATA)
             != CSIMQRO_COA_WITH_FULL_DATA) partialData = CSITRUE;
      }
      break;

      case CSIMQFB_COD:
      if ((dataOptions & (CSIMQRO_COD_WITH_DATA | CSIMQRO_COA_WITH_DATA))!=0L)
      {
         reportType  = CSIMQRO_COD_WITH_FULL_DATA        |
                       CSIMQRO_EXPIRATION_WITH_FULL_DATA |
                       CSIMQRO_EXCEPTION_WITH_FULL_DATA; 
         restoreFile = CSITRUE;
         if ((reportOptions & CSIMQRO_COD_WITH_FULL_DATA)
             != CSIMQRO_COD_WITH_FULL_DATA) partialData = CSITRUE;
      }
      else /* OR 3864: Removes temporary file generated by the asynch. LI */
      {
         deleteFile = CSITRUE;
      }
      break;

      case CSIMQFB_EXPIRATION:
/*
      if ((dataOptions & CSIMQRO_EXPIRATION_WITH_DATA)!=0L)
      {
         reportType  = CSIMQRO_COD_WITH_FULL_DATA        |
                       CSIMQRO_EXPIRATION_WITH_FULL_DATA |
                       CSIMQRO_EXCEPTION_WITH_FULL_DATA; 
         restoreFile = CSITRUE;
         if ((reportOptions & CSIMQRO_EXPIRATION_WITH_FULL_DATA)
             != CSIMQRO_EXPIRATION_WITH_FULL_DATA) partialData = CSITRUE;
      }
*/
      deleteFile = CSITRUE;
      break;

      default:
/*
      if ((dataOptions & CSIMQRO_EXCEPTION_WITH_DATA)!=0L)
      {
         reportType  = CSIMQRO_COD_WITH_FULL_DATA        |
                       CSIMQRO_EXPIRATION_WITH_FULL_DATA |
                       CSIMQRO_EXCEPTION_WITH_FULL_DATA; 
         restoreFile = CSITRUE;
         if ((reportOptions & CSIMQRO_EXCEPTION_WITH_FULL_DATA)
             != CSIMQRO_EXCEPTION_WITH_FULL_DATA) partialData = CSITRUE;
      }
*/
      deleteFile = CSITRUE;
      break;
   }

   /** BUILD THE FILE NAME, FROM THE UNIQUE ID **/
   /*********************************************/

   tmpName_P = li_BuildFileName(uniqueId, tmpFileName);
   if (NULL == tmpName_P)
   {
      status = LI_ERR_NO_FILE_NAME;
   }
   else
   { /* Building link name */
      if (strlen(tmpName_P) < sizeof(CSICHAR256)-2)
      {
         strcpy(tmpLinkName, tmpFileName);
         strcat(tmpLinkName, "=");
      }
      else
      {
         status = LI_ERR_NO_FILE_NAME;
      } 
   }

   /** IS THE FILE THERE ? **/
   /*************************/

   if (LI_ERR_NONE == status)
   {
      struct stat tmpBuffer;

      if (-1==stat(tmpFileName, &tmpBuffer)) status = LI_ERR_NO_FILE_NAME;
   }
      
   /** READ AND UPDATE REPORT FLAGS IN THE FILE **/
   /**********************************************/

   if ( (LI_ERR_NONE == status)   &&
        (restoreFile == CSITRUE)  &&
        (deleteFile  == CSIFALSE) )
   {  /* Open da file */
      ret = OS_storage_open(tmpFileName, OS_O_RDWR, &fileDesc, 
                            returnCode_P, reasonCode_P);
      if (CSIFAILED == ret) status = *reasonCode_P;
      else
      {
         /* Set a blocking write-lock on the file */

         struct flock lock;

         lock.l_type   = F_WRLCK;
         lock.l_start  = 0;
         lock.l_whence = SEEK_SET;
         lock.l_len    = 0;

         do { ret = fcntl(fileDesc, F_SETLKW, &lock); }
         while ( (ret == -1) && (errno == EINTR) );

         if (ret == -1) status = LI_ERR_INTERNAL;
         else
         {
            /* Locate the cursor at the right byte in the file */
            ret = OS_storage_seek(fileDesc, - sizeof(CSILONG), OS_SEEK_END,
                                  returnCode_P, reasonCode_P);
            if (CSIFAILED == ret) status = *reasonCode_P;
            else
            {
               /* Read the report flags */
               length = sizeof(CSILONG);
               ret = OS_storage_read(fileDesc, &length, &reportFlags,
                                     returnCode_P, reasonCode_P);
               if (CSIFAILED == ret) status = *reasonCode_P;
               else
               {  /* Updating the report flags */
                  reportFlags &= ~reportType;

/* 0R 3864: do not restore file if COD is not required by the application */
                  if ((reportFeedback == CSIMQFB_COD) &&
                      ((dataOptions & CSIMQRO_COD_WITH_DATA)==0L))
                  {
                  restoreFile = CSIFALSE;
                  }

                  if (reportFlags == 0L) deleteFile = CSITRUE;
                  else
                  { /* Update the report flags in the file */
                     ret = OS_storage_seek(fileDesc, -sizeof(CSILONG),
                           OS_SEEK_END, returnCode_P, reasonCode_P);
                     if (CSIFAILED == ret) status = *reasonCode_P;
                     else
                     {
                        length = sizeof(CSILONG);
                        ret = OS_storage_write(fileDesc, &length,
                              &reportFlags, returnCode_P, reasonCode_P);
                     }
                  }
               }
            }
         }
   /** LINK FILE IF NEEDED **/
   /*************************/

         if ((LI_ERR_NONE == status) &&
             (restoreFile == CSITRUE) && (deleteFile == CSIFALSE))
         {
            /* Create a hard link on the file */
            ret = link(tmpFileName, tmpLinkName);
/* OR 3864: link returns a value of -1 when an error occurs, not 1 */
            if (ret == -1) status = LI_ERR_INTERNAL;
            else
            {
               tmpName_P = tmpLinkName;
            }
         }
/* OR 3864: close the file after link */
         ret=OS_storage_close(fileDesc, OS_C_CLOSE, returnCode_P, reasonCode_P);
      }
   }

   /** RESTORE FILE IF NEEDED **/
   /****************************/

   if ((LI_ERR_NONE == status) &&
        (restoreFile == CSITRUE) && (deleteFile == CSITRUE))
   {  /* We use the file */
         tmpName_P = tmpFileName;
   }

   /** RESTORE FILE OR LINK **/
   /**************************/
   /* Note: we will restore either the file itself, or the hard link */

   if (LI_ERR_NONE == status)
   {
      if (restoreFile == CSITRUE)
      {
         ret = OS_storage_getsize(tmpName_P, &length,
               returnCode_P, reasonCode_P);
         length -= sizeof(CSILONG); /* Remember the flags */
         if ( (CSISUCCESS == ret) && (length > 0L) )
         {
            CSI_DD_FLAGS_SET_STORAGE( dataDesc_P, CSI_DD_FLAGS_DATA_BY_FILE );

            dataDesc_P->DataLength = ( (partialData == CSITRUE) ? 
                                   GCN_MIN(CSI_PARTIAL_DATA_LENGTH, length) :
                                       length);

            strcpy(dataDesc_P->FileName, tmpName_P);

            EVR_TRACELOG(EVR_DEBUG, (LI_TRC_ORIGINAL_DATA_RESTORED,
                         dataDesc_P->DataLength, tmpFileName));

           /* The file (or link) may be deleted by the rap after emission */
           CSI_DD_FLAGS_NO_KEEP_FILE(dataDesc_P);
         }
      }
      else
      { /* Delete the file ? */
         if (deleteFile == CSITRUE)
         {
            ret = OS_storage_delete(tmpName_P, returnCode_P, reasonCode_P);
            if (CSISUCCESS == ret)
               EVR_TRACELOG(EVR_DEBUG, (LI_TRC_ORIGINAL_DATA_DELETED,
                            tmpFileName));
         }
      }
   }

   /** THE END **/
   /*************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_INT_LOW_FUNCTION_EXIT("li_RestoreOriginalData", 
                               *returnCode_P, *reasonCode_P);
   return(*returnCode_P);
}


/**********************************************************/
/** A SMALL FUNCTION TO BUILD A NAME USING THE UNIQUE ID **/
/**********************************************************/
/*
* @LOCALFUN li_BuildFileName
*
* @DESCRIPTION Builds a file name for temporary data storage
*
* @ABSTRACT
* This function builds the name of a temporary file, to be used
* by <@LOCALFUN li_SaveOriginalData> and <@LOCALFUN li_RestoreOriginalData>.
*
* @RETURNVALUE CSICHAR*
* Either: a NULL pointer => Something went wrong, somehow...
*     or: a pointer to the 'fileName' variable. 
*
* @PARAMETER
* IN     CSICHAR48 uniqueId
*        The Unique Identifier of the message. 47 characters at most.
*        It will be used as the relative file name.
*
* OUT    CSICHAR256 fileName
*        The resulting filename. It is build using the unique id as the
*        relative file name, the path of the directory is built with the
*        environment variable CCNWKD. It is actually a subdirectory
*        of the temporary files storage area (currently
*        $CCNWKD/ccnwkd/tmp). It is hence $CCNWKD/ccnwkd/tmp/li/...
*
* @SEEALSO
* <@LOCALFUN li_SaveOriginalData>
* <@LOCALFUN li_RestoreOriginalData>
*/
CSICHAR*  li_BuildFileName(
IN        CSICHAR48    uniqueId,
OUT       CSICHAR256   fileName
)
{
   /** LOCAL VARIABLES DECLARATION **/
   /*********************************/

   CSILONG    ret        = CSISUCCESS;
   CSICHAR*   tmpName_P  = NULL;

   CSILONG    returnCode = CSISUCCESS;
   CSILONG    reasonCode = LI_ERR_NONE;

   CSICHAR256 tmpPath    = {""};

   TRACE_INT_LOW_FUNCTION_ENTRY("li_BuildFileName");

   /** GET TMP PATH + LI SUB DIR **/
   /*******************************/

   ret = gcn_GetTmpPath(CSI_ENVPATH, LI_WKPATH, tmpPath, 
                        &returnCode, &reasonCode);
   if (CSISUCCESS == ret)
   {
      /** Concat tmpPath and uniqueId **/
      /*********************************/

      if ( (strlen(tmpPath) + strlen(uniqueId)) < (sizeof(CSICHAR256) - 1) )
      {
         strcpy(fileName, tmpPath);
         strcat(fileName, uniqueId);
         tmpName_P = fileName;
      }
   }

   /** THE END **/
   /*************/

   TRACE_INT_LOW_FUNCTION_EXIT("li_BuildFileName", returnCode, reasonCode);
   return (tmpName_P);
}


/**********************************************************/
/*** CHECKING THAT A QUEUE DEPENDS ON THE LOCAL GATEWAY ***/
/**********************************************************/
/*
* @LOCALFUN li_IsQueueLocal
*
* @DESCRIPTION Checks whether a queue belongs to the local gateway.
*
* @ABSTRACT
* This function returns a boolean value indicating whether the
* given queue name is local, or remote (or invalid). This is performed
* by comparing the name of the gateway associated with the queue with
* the local gateway name. Relies on <@FUNCTION rou_GetTargetGatewayName>
* and <@FUNCTION CDX_isLocalGateway>.
*
* @RETURNVALUE CSIBOOLEAN
* CSITRUE means the queue is local, CSIFALSE that is is either remote or
* invalid.
*
* @PARAMETER
* IN     LI_MSGCONTEXT *li_MsgContext_P   The message context
* IN     CSICHAR48     queueName          The checked queue name
*
* @SEEALSO
* <@LOCALFUN li_AsyncProc>
*/
CSIBOOLEAN   li_IsQueueLocal(
IN           LI_MSGCONTEXT   *li_MsgContext_P,
IN           CSICHAR48       queueName
)
{
   /** LOCAL VARIABLES DECLARATION **/
   /*********************************/

   CSILONG   returnCode   = CSISUCCESS;
   CSILONG   reasonCode   = LI_ERR_NONE;

   CSILONG   ret          = CSISUCCESS;

   CSICHAR256 gatewayName;

   CSIBOOLEAN isLocal     = CSIFALSE;

   TRACE_INT_LOW_FUNCTION_ENTRY("li_IsQueueLocal");

   /** GET GATEWAY ASSOCIATED WITH QUEUE NAME **/
   /********************************************/

   ret = rou_GetTargetGatewayName(queueName
                                 ,li_MsgContext_P->RunMode
                                 ,CSI_ASYNCHRON
                                 ,FALSE
                                 ,&gatewayName
                                 ,&returnCode
                                 ,&reasonCode
                                 );

   /** THEN COMPARE THE RESULT WITH THE LOCAL GATEWAY **/
   /****************************************************/

   if (CSISUCCESS == ret)
   {
      ret = CDX_isLocalGateway(gatewayName, &isLocal,
                               &returnCode, &reasonCode);
   }

   /** THE END **/
   /*************/

   if (CSIFAILED == ret) isLocal = CSIFALSE;

   TRACE_INT_LOW_FUNCTION_EXIT("li_IsQueueLocal", returnCode, reasonCode);
   return(isLocal);
}


/** MACRO-LIKE FUNCTION TO PERFORM A CROSS-CHECK (REPORT w/ DATA) ***/
/********************************************************************/
/* Reminder: in case a report with data is required, the LI keeps   */
/* --------- a copy of the data in a temporary file. But, the       */
/*           message being asynchronous, a report must be generated */
/*           for this file to be deleted or restored. The received  */
/*           report may be required or not by the application, thus */
/*           it may be either transmitted or dismissed. This little */
/*           function is just such a verification.                  */
/********************************************************************/

CSIBOOLEAN li_ShouldReportBeDismissed(
IN         CSILONG    report,
IN         CSILONG    feedback
)
{
   CSIBOOLEAN result = CSIFALSE;

   switch (feedback)
   {
      case CSIMQFB_COA:
         if ( (report & CSIMQRO_COA) == 0x00000000L ) result = CSITRUE;
         break;

      case CSIMQFB_COD:
         if ( (report & CSIMQRO_COD) == 0x00000000L ) result = CSITRUE;
         break;

      case CSIMQFB_EXPIRATION:
         if ( (report & CSIMQRO_EXPIRATION) == 0x00000000L) result = CSITRUE;
         break;

      default:
         if ( (report & CSIMQRO_EXCEPTION) == 0x00000000L) result = CSITRUE;
         break;
   }

   return(result);
}


/**********************************************************/
/*** FUNCTION THAT ALLOWS TO CALL THE ALI ITSELF IF     ***/
/*** ANY MESSAGES IN THE QUEUE WHILE BOOTING            ***/
/**********************************************************/
/*
* @FUNCTION ALI_checkInputQ
* 
* @DESCRIPTION Checks if messages are available in an input queue.
* 
* @ABSTRACT
* The ALI_checkInputQ function checks if messages are present in the ALI input
* queue. If messages are present, this function sends triggering events to 
* it-self or other ALI processes and then ends. 
* This function is called by the ALI processes during their
* initialisation phase. The events sent by this function allow the ALI to 
* terminate its start and to handle the messages in its input queue later.
*
* It acts as follows :
* - opens the input queue.
* - gets the following attributes of the queue, using the MQINQ function:
*   - the maximum number of ALI processes allowed for this queue.
*   - the number of messages present in the queue.
* - if messages are present in the queue:
*   - calls the ALI service (using the <@FUNCTION CSI_acall> function).
*     (the number of "calls" performed depends on the number of messages
*      present in the queue and the maximum number of processes allowed).
* - closes the input queue.
*
*
* @RETURNVALUE CSILONG
* ALI_checkInputQ returns CSISUCCESS upon success, otherwise, it returns 
* CSIFAILED and a reason code to indicate the error condition.
*
* @PARAMETER
*
* IN  CSIMQHCONN QMgrHandle  Queue Manager Handle.
*
* IN  CSICHAR48  inputQName  Input queue name.
*
* IN  CSICHAR48  svcName     Service to call if messages are present in the
*                            queue.
*
* OUT CSILONG   *returnCodeP The return code of the function
*                            (CSISUCCESS or CSIFAILED).
*
* OUT CSILONG   *reasonCodeP In case of failure, this reason code is an
*                            error code.
*
* @ERRORCODE
*
*/

CSILONG ALI_checkInputQ(
IN   CSIMQHCONN QMgrHandle,
IN   CSICHAR48  inputQName,
IN   CSICHAR48  svcName,
OUT  CSILONG   *returnCodeP,
OUT  CSILONG   *reasonCodeP
)
{
 /* Return and reason codes */
 CSILONG       retCode, reaCode;

 /* Variables used by the CSI_mq_xx functions */
 CSILONG     openOpts;                           /* queue open options */
 CSIMQOD     inputQObjDescS = {CSIMQOD_DEFAULT}; /* queue object descriptor */
 CSIMQHOBJ   inputQHandle;
 MQLONG      attrSelector[ALI_MQINQ_REQUESTED_ATTR_MXNUM];
 MQLONG      attrValues[ALI_MQINQ_REQUESTED_INT_ATTR_MXNUM];
 MQCHAR64    triggerData;
 CSILONG     mqReturnCode, mqReasonCode;

 CSILONG     processMaxNumber;

 /* Variables used by the CSI_acall function */
 CSILONG     acallFlags      = CSITPNOREPLY;
 CSIQOS      acallQos        = {CSIQOS_DEFAULT};
 CSILONG     callDesc;
 CSIPDU     *acallDataP;

 TRACE_FUNCTION_ENTRY("ALI_checkInputQ");

 EVR_TRACELOG(EVR_PROCESS,(LI_TRC_CHECK_INPUT_QUEUE,
              inputQName));

 /* Initialise the return and reason codes. */
 *returnCodeP = CSISUCCESS;
 *reasonCodeP = LI_ERR_NONE;

 /* ---------------------------------------------------------------
  * Open the input queue.
  */
 openOpts = CSIMQOO_INQUIRE +
            CSIMQOO_FAIL_IF_QUIESCING;

 /* Set the name of the input queue name in the object descrptor. */
 strcpy(inputQObjDescS.ObjectName, inputQName);

 retCode = CSI_mq_open(QMgrHandle,
                       &inputQObjDescS,
                       openOpts,
                       &inputQHandle,
                       &retCode,
                       &reaCode);

 if ((retCode == CSIFAILED) || (retCode == CSIWARNING))
 {
   EVR_ERRORLOG(EVR_CRITICAL,(LI_MSG_MQ_OPEN_FAILED,
                "Input", retCode, reaCode));

   *returnCodeP = retCode = CSIFAILED;
 }
 else
 {
   /* Prepare attribute selector array for MQ inquire call.
      Requested attributes are:
       - Current depth: Number of messages present in the queue.
       - Trigger data : Maximum number of ALI processes supported.
      The first requested attribute is of type integer, the second one of
      type string. */

   attrSelector[ALI_MQINQ_QUEUE_DEPTH_ATTR_INDEX] = MQIA_CURRENT_Q_DEPTH;
   attrSelector[ALI_MQINQ_TRIGGER_DATA_ATTR_INDEX] = MQCA_TRIGGER_DATA;

   /* -----------------------------------------------------------------------
    * Get the maximum number of ATA processes and the number of messages
    * present in the queue.
    */ 

   MQINQ(QMgrHandle, inputQHandle,
         ALI_MQINQ_REQUESTED_ATTR_MXNUM, attrSelector,
         ALI_MQINQ_REQUESTED_INT_ATTR_MXNUM, attrValues,
         MQ_TRIGGER_DATA_LENGTH, triggerData,
         &mqReturnCode, &mqReasonCode);

   if (mqReturnCode != MQCC_OK)
   {
     /* Failed to inquire on ALI queue. */
     EVR_ERRORLOG (EVR_MINOR, (LI_MSG_MQ_INQUIRE_FAILED,
                              inputQName, mqReasonCode));
     /* Don't report this error and continue. */

   }
   else
   {
     /* Get the maximum number of ALI processes. */
     processMaxNumber = atol(triggerData);
     if (processMaxNumber == 0L)
     {
       processMaxNumber = 1L;
     }

     EVR_TRACELOG(EVR_DEBUG, (LI_TRC_MQINQ_RESULT,
                  processMaxNumber,
                  attrValues[ALI_MQINQ_QUEUE_DEPTH_ATTR_INDEX]));

     /* If there are messages present in the queue, at least one service
        call is send. The number of service calls depends on the number of
        messages present in the queue, the maximum number of ALI processes
        and the number of messages one process can handle. */
     if (attrValues[ALI_MQINQ_QUEUE_DEPTH_ATTR_INDEX] != 0)
     {
       /* ------------------------------------------------------------------
        * Allocate a data descriptor in order to perform the call.
        * ------------------------------------------------------------------ */
       retCode = gcn_AllocDataDesc(0L, &acallDataP, &retCode, &reaCode);
  
       if (retCode == CSIFAILED)
       {
         EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_GCN_ERROR,
                      "gcn_AllocDataDesc", reaCode));
  
         *returnCodeP = CSIFAILED;
       }
       else
       {

       /*********************************************************************/
       /*  This differs from the ATR behaviour's. Each LI calls only once   */
       /*  its own service. Each ALI bootting it will do it. Next 'for'     */
       /*  is then commented.                                               */
       /*********************************************************************/

       /* First build the message trace information to warn the recipient
          that this is a triggering service call. */

       (void)gcn_BuildTraceInfo(CSI_TRIGGER_EVENT, NULL, inputQName,
                                    NULL, NULL, acallDataP,
                                    &retCode, &reaCode);
       if (retCode != CSISUCCESS)
       {
         /* Failed to insert the trace information PDU in the data
            descriptor. */
         EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_GCN_ERROR,
                      "gcn_BuildTraceInfo", reaCode));
         *returnCodeP = CSIFAILED;
       }
       else
       {

         /* ----------------------------------------------------------------
          * Call the ALI service.
          * ---------------------------------------------------------------- */
         CSI_acall(svcName, (CSIDD *) acallDataP, acallFlags,
                   &acallQos, &callDesc, &retCode, &reaCode);

         if (retCode != CSISUCCESS)
         {
           EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_CALL_FAILED,
                        svcName, reaCode));
           *returnCodeP = CSIFAILED;
         }

         gcn_ResetDataDesc(acallDataP, &retCode, &reaCode);

       }

         /* Release the allocated data descriptor. */
         retCode = gcn_FreeDataDesc(&acallDataP,
                                    &retCode,
                                    &reaCode);

         if (retCode == CSIFAILED)
         {
           EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_GCN_ERROR,
                        "gcn_FreeDataDesc", reaCode));

           /* Don't report this error and continue. */
         }
       }
     }
   }


   /* -----------------------------------------------------------------
    * Close the ALI input queue.
    */

   retCode = CSI_mq_close(QMgrHandle,
                          &(inputQHandle),

                          CSIMQCO_NONE,
                          &retCode,
                          &reaCode);

   if (retCode == CSIFAILED)
   {
     EVR_ERRORLOG(EVR_UNDETERMINATE, (LI_MSG_CSI_ERROR,
                  "CSI_mq_close", reaCode));
     /* Don't report this error and continue. */
   }
 }

 if (*returnCodeP == CSIFAILED)
 {
   *reasonCodeP = LI_ERR_INTERNAL;
 }

 TRACE_FUNCTION_EXIT("ALI_checkInputQ", *returnCodeP, *reasonCodeP);

 return(*returnCodeP);
}
/*****************************************************************************/
